#ifndef TEMPLATE_TREE
#define TEMPLATE_TREE

//necessary for callback functions
#include <windows.h>
#include <windef.h>

//type is the class of the object being stored and comp is the class
//that is used to compare for a item when calling Delete
//or Find.
template<class type, class comp=type>
class CTree
{
public:
  typedef void (FAR WINAPI *ENUMTREEPROC)(type * pItem, void * pData);

private:
  class CNode
  {
  public:
    CNode * Left;
    CNode * Right;
    type * Item;

    CNode(type * i) : Left(NULL), Right(NULL), Item(i) {}
    
    ~CNode()
    {
    }

    void InFixCallback(ENUMTREEPROC proc, void * pData)
    {
      if (NULL!=Left)
        Left->InFixCallback(proc, pData);
      proc(this->Item, pData);
      if (NULL!=Right)
        Right->InFixCallback(proc, pData);
    }

    void Cleanup(bool DeleteItem)
    {
      if ((DeleteItem==TRUE) && (NULL!=Item))
      {
        delete Item;
        Item=NULL;
      }

      if (NULL!=Left)
      {
        Left->Cleanup(DeleteItem);
        delete Left;
        Left=NULL;
      }

      if (NULL!=Right)
      {
        Right->Cleanup(DeleteItem);
        delete Right;
        Right=NULL;
      }
    }

    void Insert(CNode * NewNode)
    {
      if (*NewNode->Item<*Item)
      {
        if (NULL==Left)
          Left=NewNode;
        else
          Left->Insert(NewNode);
      }
      else
      {
        if (NULL==Right)
          Right=NewNode;
        else
          Right->Insert(NewNode);
      }
    }

    void Zig()
    {
      type * itmp=Item;
      Item=Right->Item;
      Right->Item=itmp;
      CNode * r=Right->Right;
      Right->Right=Right->Left;
      Right->Left=Left;
      Left=Right;
      Right=r;
    }

    void Zag()
    {
      type * itmp=Item;
      Item=Left->Item;
      Left->Item=itmp;
      CNode * l=Left->Left;
      Left->Left=Left->Right;
      Left->Right=Right;
      Right=Left;
      Left=l;
    }

    void Splay(comp SearchItem)
    {
      if (*Item==SearchItem)
        return;

      if (*Item<SearchItem)
      {
        if (NULL!=Right)
        {
          Right->Splay(SearchItem);
          Zig();
        }
      }
      else
      {
        if (NULL!=Left)
        {
          Left->Splay(SearchItem);
          Zag();
        }
      }
    }

    void DeleteMin()
    {
      if (Left==NULL)
        return;
      Left->DeleteMin();
      Zag();
    }

    void DeleteMax()
    {
      if (Right==NULL)
        return;
      Right->DeleteMax();
      Zig();
    }

  };

  CNode * Head;
  int Count;
  bool DeleteItem;

  void Splay(CNode *& Start, comp SearchItem)
  {
    Start->Splay(SearchItem);
  }


public:
  CTree(bool DeleteOnDestroy)
    : Count(0), Head(NULL), DeleteItem(DeleteOnDestroy) {}
  CTree() : Count(0), Head(NULL), DeleteItem(true) {}
  ~CTree()
  {
    if (NULL!=Head)
    {
      Head->Cleanup(DeleteItem);
      delete Head;
      Head=NULL;
    }
  }

  int ItemCount()
  {
    return Count;
  }

  void Insert(type * Item)
  {
    Count++;
    CNode * NewNode = new CNode(Item);
    if (NULL==Head)
      Head=NewNode;
    else
      Head->Insert(NewNode);
  }

  type * Find(comp SearchItem)
  {
    if (NULL==Head)
      return NULL;

    Splay(Head, SearchItem);

    if (*Head->Item==SearchItem)
      return Head->Item;
    else
      return NULL;
  }

  type * DeleteMin()
  {
    if (NULL==Head)
      return NULL;
    Head->DeleteMin();
    CNode * NewHead=Head->Right;
    type * ret=Head->Item;
    delete Head;
    Head=NewHead;
    Count--;
    return ret;
  }

  type * DeleteMax()
  {
    if (NULL==Head)
      return NULL;
    Head->DeleteMax();
    CNode * NewHead=Head->Left;
    type * ret=Head->Item;
    delete Head;
    Head=NewHead;
    Count--;
    return ret;
  }

  void SetDelete(bool Delete)
  {
    DeleteItem=Delete;
  }

  bool GetDelete()
  {
    return DeleteItem;
  }

  void InFixCallback(ENUMTREEPROC proc, void * pData)
  {
    if (NULL!=Head)
      Head->InFixCallback(proc, pData);
  }

  type * Delete(comp item)
  {
    if (NULL==Head)
      return NULL;

    Splay(Head, item);
    
    if (*Head->Item==item)
    {
      CNode * OldHead=Head;
      type * ret=OldHead->Item;

      if (Head->Left)
      {
        if (Head->Right)
        {
          Splay(Head->Left, item);
          Head->Left->Insert(Head->Right);
        }
        Head=Head->Left;
      }
      else
        Head=Head->Right;

      Count--;
      delete OldHead;
      return ret;
    }
    return NULL;
  }
};

#endif